@@ -1,3 +1,4 @@
+from logging import exception
 import mmap
 import os
 import random
@@ -84,9 +85,9 @@ class DetectType:
         return "STRING"
 
 
-    def __infer_value_type(self, value, index, schema):     
-        
-        if value not in schema[index]["values"].keys():
+    def __infer_value_type(self, value, index, schema, values_type): 
+
+        if value not in values_type.keys():
             
             local_type = self.__get_local_type(value)
             
@@ -103,21 +104,26 @@ class DetectType:
                     _type = local_type
             else:
                 _type = local_type
-            
-            schema[index]["values"][value] = { "cnt": 1,"_type": _type}
-        
+
+            values_type[value] = _type
+
+            if values_type[value] not in schema[index]["types_found"].keys():
+                schema[index]["types_found"][values_type[value]] = { "cnt": 1}
+            else:
+                schema[index]["types_found"][values_type[value]]["cnt"] += 1
         else:
-            schema[index]["values"][value]["cnt"] += 1
-            
+            if values_type[value] not in schema[index]["types_found"].keys():
+                schema[index]["types_found"][values_type[value]] = { "cnt": 1}
+            else:
+                schema[index]["types_found"][values_type[value]]["cnt"] += 1
+        
 
     def execute(self, records, schema):
-
+        values_type = {}
         for record in records:
-            values = record.rstrip().split(self.sep)            
+            values = record.rstrip().split(self.sep)              
             for index, value in enumerate(values):
-                self.__infer_value_type(value[0:self.max_length], index, schema)
-        
-        return schema
+                self.__infer_value_type(value[0:self.max_length], index, schema, values_type)
 
 
 class Parallel:
@@ -127,15 +133,28 @@ class Parallel:
 
     
     def execute(self, records, x, obj, d_schema):
-        return obj.execute(records, d_schema)        
-
+        obj.execute(records, d_schema)
+        return d_schema
 
         
-    def parallel(self, records, obj,  d_schema, chunk_size = None):
+    def parallel(self, records, obj,  d_schema):
         
-        cpus = (mp.cpu_count() - 1)
-        if chunk_size is None:
-            chunk_size = len(records) // cpus                
+
+
+        cpus = (mp.cpu_count() - 2)
+
+        if cpus <= 0:
+            cpus = mp.cpu_count()
+        
+        chunk_size = len(records) / cpus
+
+        if chunk_size < 1:
+            cpus = int(chunk_size * 10)
+            chunk_size = 1
+        else:
+            chunk_size = round(chunk_size)
+        
+
 
         pool = mp.Pool(processes=cpus)
         
@@ -143,16 +162,12 @@ class Parallel:
         pool.close()
         pool.join()
 
-
-        outputs = [p.get() for p in results]
-
-        return outputs
-        
+        return [p.get() for p in results]
 
 
 class CsvSchemaInference:
     
-    def __init__(self, portion = 0.5, max_length = 1000, acc = 0.7, seed= 1, header= True, sep=";", conditions = {}):
+    def __init__(self, portion = 0.5, max_length = 1000, batch_size = 250000, acc = 0.7, seed= 1, header= True, sep=";", conditions = {}):
         self.portion = portion
         self.seed = seed
         self.header = header
@@ -161,6 +176,7 @@ class CsvSchemaInference:
         self.__schema = {}
         self.max_length = max_length
         self.data_types = {"STRING", "INTEGER", "FLOAT", "DATETIME", "DATE", "TIME", "TIMESTAMP", "BOOLEAN"}
+        self.batch_size = batch_size
 
         if isinstance(conditions,dict):
 
@@ -181,10 +197,10 @@ class CsvSchemaInference:
         for i in range(0, len(header)):
             self.__schema[i] = {
                 "_name": header[i].replace('"', ''),
-                "values":{
+                "types_found":{                    
                 },
                 "nullable":False,
-                "approximate_type":""
+                "type":""
             }
 
 
@@ -200,22 +216,21 @@ class CsvSchemaInference:
 
             for s_inx in range(0, len(schemas)):
 
-                v_dict = schemas[s_inx][c_inx]
+                _v = schemas[s_inx][c_inx]
 
-
-                if v_dict['nullable']:
+                if _v['nullable']:
                     self.__schema[c_inx]['nullable'] = True
 
-                for k in v_dict['values']:
 
-                    if k not in self.__schema[c_inx]['values']:
+                for k in _v['types_found']:
 
-                        self.__schema[c_inx]['values'][k] = { 
-                                    "cnt": v_dict['values'][k]['cnt'],
-                                    "_type": v_dict['values'][k]['_type']
+                    if k not in self.__schema[c_inx]['types_found'].keys():
+
+                        self.__schema[c_inx]['types_found'][k] = { 
+                                    "cnt": _v['types_found'][k]['cnt']
                                     }
                     else:
-                        self.__schema[c_inx]['values'][k]['cnt'] += v_dict['values'][k]['cnt']
+                        self.__schema[c_inx]['types_found'][k]['cnt'] += _v['types_found'][k]['cnt']
 
 
 
@@ -247,17 +262,16 @@ class CsvSchemaInference:
 
     def __approximate_types(self, acc = 0.5):
 
-        result = {}        
+        result = {}
         for c in self.__schema:
             _types = {}
             t = 0
-            for v in self.__schema[c]['values']:
-                value = self.__schema[c]['values'][v]
-                t += value['cnt']
-                if value['_type'] not in _types:
-                    _types[value['_type']] = value['cnt']
-                else:
-                    _types[value['_type']] += value['cnt']
+            for v in self.__schema[c]['types_found']:
+                t += self.__schema[c]['types_found'][v]['cnt']
+                if v not in _types.keys():
+                    _types[v] = self.__schema[c]['types_found'][v]['cnt']
+                else:                    
+                    _types[v] += self.__schema[c]['types_found'][v]['cnt']
 
             for ft in _types:                
                 _types[ft] = (_types[ft] * 100) / t
@@ -266,7 +280,7 @@ class CsvSchemaInference:
             _type = self.check_condition(_types, acc)
 
 
-            self.__schema[c]['approximate_type'] = _type
+            self.__schema[c]['type'] = _type
 
             result[c] = {
                 "name": self.__schema[c]['_name'], 
@@ -296,9 +310,9 @@ class CsvSchemaInference:
             if self.__schema[c]["_name"] in columns:
                 result[c] = {
                     "_name": self.__schema[c]["_name"],
-                    "values":self.__schema[c]["values"],
+                    "types_found":self.__schema[c]["types_found"],
                     "nullable":self.__schema[c]["nullable"],
-                    "approximate_type":self.__schema[c]["approximate_type"]
+                    "type":self.__schema[c]["type"]
                 }
         
         return result
@@ -314,22 +328,20 @@ class CsvSchemaInference:
 
                 _types = {}
                 t = 0
-                for v in self.__schema[c]['values']:
-                    value = self.__schema[c]['values'][v]
-                    t += value['cnt']
-                    if value['_type'] not in _types:
-                        _types[value['_type']] = value['cnt']
+                for v in self.__schema[c]['types_found']:
+                    t += self.__schema[c]['types_found'][v]['cnt']
+
+                    if v not in _types.keys():
+                        _types[v] = self.__schema[c]['types_found'][v]['cnt']
                     else:
-                        _types[value['_type']] += value['cnt']
+                        _types[v] += self.__schema[c]['types_found'][v]['cnt']
 
                 for ft in _types:
-                    p = (_types[ft] * 100) / t
-                    _types[ft] = p
+                     _types[ft] = (_types[ft] * 100) / t
 
-                _types
                 result[c] = {
                     "name" : self.__schema[c]['_name'],
-                    "types": _types,
+                    "types_found": _types,
                     "nullable": self.__schema[c]['nullable']
                 }
                 
@@ -341,35 +353,58 @@ class CsvSchemaInference:
 
     def run_inference(self, filename):
 
-        with open(filename, mode="r", encoding="utf8") as file_obj:
+        with open(filename, mode="r", encoding = "ISO-8859-1") as file_obj:
 
-            with mmap.mmap(file_obj.fileno(), length=0, access=mmap.ACCESS_READ) as map:
-                
-                no_lines = self.__estimate_count(filename, map)
-                portion = int(no_lines * self.portion)
-                map.seek(0)
+            with mmap.mmap(file_obj.fileno(), length=0, access=mmap.ACCESS_READ) as map_file:
+
+                less_header = 0
 
                 if self.header:
-                    self.__set_header(map.readline().decode("utf-8"))                
+                    less_header = 1
 
-                random.seed(self.seed)
+                no_lines = self.__estimate_count(filename, map_file) - less_header
+                portion = int(no_lines * self.portion)
+                map_file.seek(0)
+                
+                if self.header:
+                    self.__set_header(map_file.readline().decode("ISO-8859-1"))
 
-                records = random.sample(map.read()
-                                    .decode("utf-8")
-                                    .splitlines(),
-                                    portion)                            
-                                    
-                prl = Parallel()
-                dtype = DetectType(self.max_length, 
-                                   self.sep)
+                lines = []
+                schemas = []
+                batch_count = 0
+
+                dtype = DetectType(self.max_length, self.sep) 
+
+                
+                while batch_count < portion:
+                    
+                    batch_count += 1
+                    lines.append(map_file.readline().decode("ISO-8859-1"))
+                    
+                    if batch_count % self.batch_size == 0:
+
+                        prl = Parallel()
+                        schemas_result = prl.parallel(records = lines, obj=dtype, d_schema = self.__schema)
+
+                        for schema in schemas_result:
+                            schemas.append(schema)
+
+                        lines = []
+
+                if len(lines) > 0:
+
+                    prl = Parallel()                                      
+                    schemas_result = prl.parallel(records = lines,obj=dtype, d_schema = self.__schema)
+
+                    for schema in schemas_result:
+                        schemas.append(schema)                    
+                    
+                    del lines
+                    del batch_count
 
-                schemas = prl.parallel(records = records,
-                                      obj=dtype, 
-                                      d_schema = self.__schema, 
-                                      chunk_size = None)
 
                 #Joining schemas results
                 self.__merge_schemas(schemas)
 
-                #Approximate data types
-                return self.__approximate_types(acc = self.accuracy)
\ No newline at end of file
+        #Approximate data types
+        return self.__approximate_types(acc = self.accuracy)
\ No newline at end of file
